from .entrypoints import validate

# version compliant with https://www.python.org/dev/peps/pep-0440/
__version__ = "0.1.0"
# Don't forget to change the version number in pyproject.toml, Dockerfile, and CITATION.cff along with this one

__all__ = ["validate"]
